@@ -22,7 +22,7 @@
 
 #include <vector>
 
-#include "s2/base/logging.h"
+#include "s2/_fp_contract_off.h"  // IWYU pragma: keep
 #include "s2/s1angle.h"
 #include "s2/s1chord_angle.h"
 #include "s2/s2cell.h"
@@ -258,7 +258,7 @@ class S2ClosestPointQuery {
   // Returns a reference to the underlying S2PointIndex.
   const Index& index() const;
 
-  // Returns the query options.  Options can be modifed between queries.
+  // Returns the query options.  Options can be modified between queries.
   const Options& options() const;
   Options* mutable_options();
 
