@@ -26,9 +26,17 @@
 #include <utility>
 #include <vector>
 
-#include "s2/base/casts.h"
+#include <gmock/gmock.h>
 #include <gtest/gtest.h>
+
+#include "absl/algorithm/container.h"
+#include "absl/log/absl_log.h"
+#include "absl/log/log_streamer.h"
+#include "absl/random/bit_gen_ref.h"
+#include "absl/random/random.h"
 #include "absl/strings/string_view.h"
+
+#include "s2/base/casts.h"
 #include "s2/util/coding/coder.h"
 #include "s2/mutable_s2shape_index.h"
 #include "s2/s1angle.h"
@@ -37,12 +45,14 @@
 #include "s2/s2coder_testing.h"
 #include "s2/s2contains_point_query.h"
 #include "s2/s2error.h"
+#include "s2/s2fractal.h"
 #include "s2/s2latlng.h"
 #include "s2/s2lax_loop_shape.h"
 #include "s2/s2loop.h"
 #include "s2/s2point.h"
 #include "s2/s2pointutil.h"
 #include "s2/s2polygon.h"
+#include "s2/s2random.h"
 #include "s2/s2shape.h"
 #include "s2/s2shapeutil_contains_brute_force.h"
 #include "s2/s2shapeutil_testing.h"
@@ -50,59 +60,94 @@
 #include "s2/s2text_format.h"
 
 using absl::string_view;
+using s2coding::CodingHint;
+using s2textformat::MakePointOrDie;
 using s2textformat::MakePolygonOrDie;
 using std::make_unique;
 using std::unique_ptr;
 using std::vector;
+using testing::HasSubstr;
 
-// Verifies that EncodedS2LaxPolygonShape behaves identically to
-// S2LaxPolygonShape. Also supports testing that the encoded form is identical
-// to the re-encoded form.
-
-void TestEncodedS2LaxPolygonShape(const S2LaxPolygonShape& original) {
-  Encoder encoder;
-  original.Encode(&encoder, s2coding::CodingHint::COMPACT);
-  Decoder decoder(encoder.base(), encoder.length());
-  EncodedS2LaxPolygonShape encoded;
-  ASSERT_TRUE(encoded.Init(&decoder));
-  EXPECT_EQ(encoded.num_loops(), original.num_loops());
-  EXPECT_EQ(encoded.num_vertices(), original.num_vertices());
-  EXPECT_EQ(encoded.num_edges(), original.num_edges());
-  EXPECT_EQ(encoded.num_chains(), original.num_chains());
-  EXPECT_EQ(encoded.dimension(), original.dimension());
-  EXPECT_EQ(encoded.is_empty(), original.is_empty());
-  EXPECT_EQ(encoded.is_full(), original.is_full());
-  EXPECT_EQ(encoded.GetReferencePoint(), original.GetReferencePoint());
-  for (int i = 0; i < original.num_loops(); ++i) {
-    EXPECT_EQ(encoded.num_loop_vertices(i), original.num_loop_vertices(i));
-    EXPECT_EQ(encoded.chain(i), original.chain(i));
-    for (int j = 0; j < original.num_loop_vertices(i); ++j) {
-      EXPECT_EQ(encoded.loop_vertex(i, j), original.loop_vertex(i, j));
-      EXPECT_EQ(encoded.chain_edge(i, j), original.chain_edge(i, j));
+template <typename LaxShape>
+void TestLaxDecoding(const LaxShape& shape, const S2LaxPolygonShape& expected) {
+  EXPECT_EQ(shape.num_loops(), expected.num_loops());
+  EXPECT_EQ(shape.num_vertices(), expected.num_vertices());
+  EXPECT_EQ(shape.num_edges(), expected.num_edges());
+  EXPECT_EQ(shape.num_chains(), expected.num_chains());
+  EXPECT_EQ(shape.dimension(), expected.dimension());
+  EXPECT_EQ(shape.is_empty(), expected.is_empty());
+  EXPECT_EQ(shape.is_full(), expected.is_full());
+  EXPECT_EQ(shape.GetReferencePoint(), expected.GetReferencePoint());
+  for (int i = 0; i < expected.num_loops(); ++i) {
+    EXPECT_EQ(shape.num_loop_vertices(i), expected.num_loop_vertices(i));
+    EXPECT_EQ(shape.chain(i), expected.chain(i));
+    for (int j = 0; j < expected.num_loop_vertices(i); ++j) {
+      EXPECT_EQ(shape.loop_vertex(i, j), expected.loop_vertex(i, j));
+      EXPECT_EQ(shape.chain_edge(i, j), expected.chain_edge(i, j));
     }
   }
   // Now test all the edges in a random order in order to exercise the cases
   // involving prev_loop_.
-  vector<int> edge_ids(original.num_edges());
+  vector<int> edge_ids(expected.num_edges());
   std::iota(edge_ids.begin(), edge_ids.end(), 0);
   std::shuffle(edge_ids.begin(), edge_ids.end(), std::mt19937_64());
   for (int e : edge_ids) {
-    EXPECT_EQ(encoded.chain_position(e), original.chain_position(e));
-    EXPECT_EQ(encoded.edge(e), original.edge(e));
+    EXPECT_EQ(shape.chain_position(e), expected.chain_position(e));
+    EXPECT_EQ(shape.edge(e), expected.edge(e));
   }
+}
 
+template <typename LaxShape>
+void TestLaxRecoding(const LaxShape& shape, const Encoder& expected) {
   // Let's also test that the encoded form can be encoded, yielding the same
   // bytes as the originally encoded form.
   Encoder reencoder;
-  encoded.Encode(&reencoder, s2coding::CodingHint::COMPACT);
-  ASSERT_EQ(string_view(encoder.base(), encoder.length()),
+  shape.Encode(&reencoder, s2coding::CodingHint::COMPACT);
+  ASSERT_EQ(string_view(expected.base(), expected.length()),
             string_view(reencoder.base(), reencoder.length()));
 }
 
+// Tests encoding an S2LaxPolygonShape and decoding it through both
+// EncodedS2LaxPolygonShape and the Init() methods of S2LaxPolygonShape.
+void TestS2LaxPolygonShapeEncoding(const S2LaxPolygonShape& original) {
+  Encoder encoder;
+  original.Encode(&encoder, s2coding::CodingHint::COMPACT);
+
+  // Test EncodedS2LaxPolygonShape API.
+  {
+    Decoder decoder(encoder.base(), encoder.length());
+    EncodedS2LaxPolygonShape shape;
+    ASSERT_TRUE(shape.Init(&decoder));
+    TestLaxDecoding(shape, original);
+    TestLaxRecoding(shape, encoder);
+  }
+
+  // Test S2LaxPolygon::Init() API.
+  {
+    Decoder decoder(encoder.base(), encoder.length());
+    S2LaxPolygonShape shape;
+    ASSERT_TRUE(shape.Init(&decoder));
+    TestLaxDecoding(shape, original);
+    TestLaxRecoding(shape, encoder);
+  }
+
+  // Test S2LaxPolygon::Init(S2Error&) API.
+  {
+    Decoder decoder(encoder.base(), encoder.length());
+    S2Error error;
+    S2LaxPolygonShape shape;
+    ASSERT_TRUE(shape.Init(&decoder, error));
+    ASSERT_TRUE(error.ok());
+    TestLaxDecoding(shape, original);
+    TestLaxRecoding(shape, encoder);
+  }
+}
+
 TEST(S2LaxPolygonShape, EmptyPolygon) {
-  S2_LOG(INFO) << "sizeof(S2LaxPolygonShape) == " << sizeof(S2LaxPolygonShape);
-  S2_LOG(INFO) << "sizeof(EncodedS2LaxPolygonShape) == "
-            << sizeof(EncodedS2LaxPolygonShape);
+  ABSL_LOG(INFO) << "sizeof(S2LaxPolygonShape) == "
+                 << sizeof(S2LaxPolygonShape);
+  ABSL_LOG(INFO) << "sizeof(EncodedS2LaxPolygonShape) == "
+                 << sizeof(EncodedS2LaxPolygonShape);
 
   S2LaxPolygonShape shape((S2Polygon()));
   EXPECT_EQ(0, shape.num_loops());
@@ -113,7 +158,7 @@ TEST(S2LaxPolygonShape, EmptyPolygon) {
   EXPECT_TRUE(shape.is_empty());
   EXPECT_FALSE(shape.is_full());
   EXPECT_FALSE(shape.GetReferencePoint().contained);
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
 TEST(S2LaxPolygonShape, Move) {
@@ -125,13 +170,11 @@ TEST(S2LaxPolygonShape, Move) {
   const S2LaxPolygonShape correct(loops);
   S2LaxPolygonShape to_move(loops);
   s2testing::ExpectEqual(correct, to_move);
-  EXPECT_EQ(correct.id(), to_move.id());
 
   // Test the move constructor.
   S2LaxPolygonShape move1(std::move(to_move));
   s2testing::ExpectEqual(correct, move1);
-  TestEncodedS2LaxPolygonShape(move1);
-  EXPECT_EQ(correct.id(), move1.id());
+  TestS2LaxPolygonShapeEncoding(move1);
   ASSERT_EQ(loops.size(), move1.num_loops());
   ASSERT_EQ(6, move1.num_vertices());
   for (int i = 0; i < loops.size(); ++i) {
@@ -144,8 +187,7 @@ TEST(S2LaxPolygonShape, Move) {
   S2LaxPolygonShape move2;
   move2 = std::move(move1);
   s2testing::ExpectEqual(correct, move2);
-  TestEncodedS2LaxPolygonShape(move2);
-  EXPECT_EQ(correct.id(), move2.id());
+  TestS2LaxPolygonShapeEncoding(move2);
   ASSERT_EQ(loops.size(), move2.num_loops());
   ASSERT_EQ(6, move2.num_vertices());
   for (int i = 0; i < loops.size(); ++i) {
@@ -155,28 +197,6 @@ TEST(S2LaxPolygonShape, Move) {
   }
 }
 
-TEST(S2LaxPolygonShape, MoveFromShapeIndex) {
-  // Construct an index containing shapes to be moved.
-  const vector<S2LaxPolygonShape::Loop> loops = {
-      s2textformat::ParsePointsOrDie("0:0, 0:3, 3:3"),
-      s2textformat::ParsePointsOrDie("1:1, 2:2, 1:2")};
-  MutableS2ShapeIndex index;
-  index.Add(make_unique<S2LaxPolygonShape>(loops));
-  index.Add(make_unique<S2LaxPolygonShape>(loops));
-  ASSERT_EQ(index.num_shape_ids(), 2);
-
-  // Verify that the move constructor moves the id.
-  S2LaxPolygonShape& shape0 = *down_cast<S2LaxPolygonShape*>(index.shape(0));
-  S2LaxPolygonShape moved_shape0 = std::move(shape0);
-  EXPECT_EQ(moved_shape0.id(), 0);
-
-  // Verify that the move-assignment operator moves the id.
-  S2LaxPolygonShape& shape1 = *down_cast<S2LaxPolygonShape*>(index.shape(1));
-  S2LaxPolygonShape moved_shape1;
-  moved_shape1 = std::move(shape1);
-  EXPECT_EQ(moved_shape1.id(), 1);
-}
-
 TEST(S2LaxPolygonShape, FullPolygon) {
   S2LaxPolygonShape shape(S2Polygon(s2textformat::MakeLoopOrDie("full")));
   EXPECT_EQ(1, shape.num_loops());
@@ -187,7 +207,7 @@ TEST(S2LaxPolygonShape, FullPolygon) {
   EXPECT_FALSE(shape.is_empty());
   EXPECT_TRUE(shape.is_full());
   EXPECT_TRUE(shape.GetReferencePoint().contained);
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
 TEST(S2LaxPolygonShape, SingleVertexPolygon) {
@@ -210,7 +230,7 @@ TEST(S2LaxPolygonShape, SingleVertexPolygon) {
   EXPECT_FALSE(shape.is_empty());
   EXPECT_FALSE(shape.is_full());
   EXPECT_FALSE(shape.GetReferencePoint().contained);
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
 TEST(S2LaxPolygonShape, SingleLoopPolygon) {
@@ -237,7 +257,7 @@ TEST(S2LaxPolygonShape, SingleLoopPolygon) {
   EXPECT_FALSE(shape.is_empty());
   EXPECT_FALSE(shape.is_full());
   EXPECT_FALSE(s2shapeutil::ContainsBruteForce(shape, S2::Origin()));
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
 TEST(S2LaxPolygonShape, MultiLoopPolygon) {
@@ -270,7 +290,7 @@ TEST(S2LaxPolygonShape, MultiLoopPolygon) {
   EXPECT_FALSE(shape.is_empty());
   EXPECT_FALSE(shape.is_full());
   EXPECT_FALSE(s2shapeutil::ContainsBruteForce(shape, S2::Origin()));
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
 TEST(S2LaxPolygonShape, MultiLoopS2Polygon) {
@@ -279,9 +299,9 @@ TEST(S2LaxPolygonShape, MultiLoopS2Polygon) {
   auto polygon = MakePolygonOrDie("0:0, 0:3, 3:3; 1:1, 1:2, 2:2");
   S2LaxPolygonShape shape(*polygon);
   for (int i = 0; i < polygon->num_loops(); ++i) {
-    S2Loop* loop = polygon->loop(i);
-    for (int j = 0; j < loop->num_vertices(); ++j) {
-      EXPECT_EQ(loop->oriented_vertex(j),
+    const S2Loop& loop = *polygon->loop(i);
+    for (int j = 0; j < loop.num_vertices(); ++j) {
+      EXPECT_EQ(loop.oriented_vertex(j),
                 shape.loop_vertex(i, j));
     }
   }
@@ -290,12 +310,13 @@ TEST(S2LaxPolygonShape, MultiLoopS2Polygon) {
 TEST(S2LaxPolygonShape, ManyLoopPolygon) {
   // Test a polygon with enough loops so that binary search is used to find
   // the loop containing a given edge.
+  absl::BitGen bitgen(S2Testing::MakeTaggedSeedSeq(
+      "MANY_LOOP_POLYGON", absl::LogInfoStreamer(__FILE__, __LINE__).stream()));
   vector<vector<S2Point>> loops;
   for (int i = 0; i < 100; ++i) {
     S2Point center(S2LatLng::FromDegrees(0, i));
-    loops.push_back(
-        S2Testing::MakeRegularPoints(center, S1Angle::Degrees(0.1),
-                                     S2Testing::rnd.Uniform(3)));
+    loops.push_back(S2Testing::MakeRegularPoints(center, S1Angle::Degrees(0.1),
+                                                 absl::Uniform(bitgen, 0, 3)));
   }
   S2LaxPolygonShape shape(loops);
   EXPECT_EQ(loops.size(), shape.num_loops());
@@ -326,17 +347,13 @@ TEST(S2LaxPolygonShape, ManyLoopPolygon) {
     }
   }
   std::shuffle(edges.begin(), edges.end(), std::mt19937_64());
-  // TODO(user,b/210097200): Use structured bindings when we require
-  // C++17 in opensource.
-  for (const auto t : edges) {
-    int e, i, j;
-    std::tie(e, i, j) = t;
+  for (const auto [e, i, j] : edges) {
     EXPECT_EQ(shape.chain_position(e), S2Shape::ChainPosition(i, j));
     auto v0 = loops[i][j];
     auto v1 = loops[i][(j + 1) % loops[i].size()];
     EXPECT_EQ(shape.edge(e), S2Shape::Edge(v0, v1));
   }
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
 TEST(S2LaxPolygonShape, DegenerateLoops) {
@@ -346,7 +363,7 @@ TEST(S2LaxPolygonShape, DegenerateLoops) {
       s2textformat::ParsePointsOrDie("5:5, 6:6")};
   S2LaxPolygonShape shape(loops);
   EXPECT_FALSE(shape.GetReferencePoint().contained);
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
 TEST(S2LaxPolygonShape, InvertedLoops) {
@@ -355,38 +372,40 @@ TEST(S2LaxPolygonShape, InvertedLoops) {
       s2textformat::ParsePointsOrDie("3:4, 3:3, 4:4")};
   S2LaxPolygonShape shape(loops);
   EXPECT_TRUE(s2shapeutil::ContainsBruteForce(shape, S2::Origin()));
-  TestEncodedS2LaxPolygonShape(shape);
+  TestS2LaxPolygonShapeEncoding(shape);
 }
 
-void CompareS2LoopToShape(const S2Loop& loop, unique_ptr<S2Shape> shape) {
+void CompareS2LoopToShape(absl::BitGenRef bitgen, const S2Loop& loop,
+                          unique_ptr<S2Shape> shape) {
   MutableS2ShapeIndex index;
   index.Add(std::move(shape));
   S2Cap cap = loop.GetCapBound();
   auto query = MakeS2ContainsPointQuery(&index);
   for (int iter = 0; iter < 100; ++iter) {
-    S2Point point = S2Testing::SamplePoint(cap);
-    EXPECT_EQ(loop.Contains(point),
-              query.ShapeContains(*index.shape(0), point));
+    S2Point point = s2random::SamplePoint(bitgen, cap);
+    EXPECT_EQ(loop.Contains(point), query.ShapeContains(0, point));
   }
 }
 
 TEST(S2LaxPolygonShape, CompareToS2Loop) {
+  absl::BitGen bitgen(S2Testing::MakeTaggedSeedSeq(
+      "COMPARE_TO_S2_LOOP", absl::LogInfoStreamer(__FILE__, __LINE__).stream()));
   for (int iter = 0; iter < 100; ++iter) {
-    S2Testing::Fractal fractal;
-    fractal.set_max_level(S2Testing::rnd.Uniform(5));
-    fractal.set_fractal_dimension(1 + S2Testing::rnd.RandDouble());
-    S2Point center = S2Testing::RandomPoint();
-    unique_ptr<S2Loop> loop(fractal.MakeLoop(
-        S2Testing::GetRandomFrameAt(center), S1Angle::Degrees(5)));
+    S2Fractal fractal(bitgen);
+    fractal.set_max_level(absl::Uniform(bitgen, 0, 5));
+    fractal.set_fractal_dimension(1 + absl::Uniform(bitgen, 0.0, 1.0));
+    S2Point center = s2random::Point(bitgen);
+    unique_ptr<S2Loop> loop(fractal.MakeLoop(s2random::FrameAt(bitgen, center),
+                                             S1Angle::Degrees(5)));
 
     // Compare S2Loop to S2LaxLoopShape.
-    CompareS2LoopToShape(*loop, make_unique<S2LaxLoopShape>(*loop));
+    CompareS2LoopToShape(bitgen, *loop, make_unique<S2LaxLoopShape>(*loop));
 
     // Compare S2Loop to S2LaxPolygonShape.
     vector<S2LaxPolygonShape::Loop> loops(
         1, vector<S2Point>(&loop->vertex(0),
                            &loop->vertex(0) + loop->num_vertices()));
-    CompareS2LoopToShape(*loop, make_unique<S2LaxPolygonShape>(loops));
+    CompareS2LoopToShape(bitgen, *loop, make_unique<S2LaxPolygonShape>(loops));
   }
 }
 
@@ -477,3 +496,47 @@ TEST(S2LaxPolygonShape, ChainVertexIteratorWorks) {
     ++chain_counter;
   }
 }
+
+std::string DecodeS2LaxPolygonShape(absl::string_view data) {
+  Decoder decoder(data.data(), data.size());
+  S2Error error;
+
+  S2LaxPolygonShape shape;
+  (void)shape.Init(&decoder, error);
+  if (!error.ok()) {
+    return std::string(error.message());
+  }
+  return {};
+}
+
+TEST(S2LaxPolygonShapeTest, InsufficientDataInEncoder) {
+  EXPECT_THAT(DecodeS2LaxPolygonShape(""), HasSubstr("Insufficient data"));
+}
+
+TEST(S2LaxPolygonShapeTest, BadVersionNumber) {
+  EXPECT_THAT(DecodeS2LaxPolygonShape("\373"), HasSubstr("Bad version number"));
+}
+
+TEST(S2LaxPolygonShapeTest, BadLoopNumber) {
+  EXPECT_THAT(DecodeS2LaxPolygonShape("\001"), HasSubstr("number of loops"));
+}
+
+TEST(S2LaxPolygonShapeTest, BadVerticesInit) {
+  EXPECT_THAT(DecodeS2LaxPolygonShape("\001\003"),
+              HasSubstr("decode vertices"));
+}
+
+TEST(S2LaxPolygonShapeTest, BadVertices) {
+  EXPECT_THAT(
+      DecodeS2LaxPolygonShape(std::string(
+          "\0014\331\227\360\360."
+          "\010\010\010\010\010\010\010\010\010\010\010\000\010\010\360\360\360"
+          "\360\360\360\360\360\360\360\360\360\360\000\251\021\021\014",
+          39)),
+      HasSubstr("Invalid exception delta outside of block size"));
+}
+
+TEST(S2LaxPolygonShapeTest, BadLoopOffsets) {
+  EXPECT_THAT(DecodeS2LaxPolygonShape(std::string("\001\225\243C\000\373", 6)),
+              HasSubstr("Failed to decode loop offsets"));
+}
