@@ -19,8 +19,10 @@
 
 #include <algorithm>
 #include <cmath>
+#include <cstdint>
 #include <iomanip>
 #include <memory>
+#include <random>
 #include <string>
 #include <utility>
 #include <vector>
@@ -32,12 +34,16 @@
 #include "absl/container/flat_hash_map.h"
 #include "absl/container/flat_hash_set.h"
 #include "absl/flags/flag.h"
+#include "absl/log/absl_check.h"
+#include "absl/log/absl_log.h"
+#include "absl/log/log_streamer.h"
+#include "absl/random/bit_gen_ref.h"
+#include "absl/random/random.h"
+#include "absl/strings/str_cat.h"
 #include "absl/strings/string_view.h"
+#include "absl/types/span.h"
 
-#include "s2/base/commandlineflags.h"
-#include "s2/base/commandlineflags_declare.h"
-#include "s2/base/logging.h"
-#include "s2/util/coding/coder.h"
+#include "s2/base/casts.h"
 #include "s2/r1interval.h"
 #include "s2/s1angle.h"
 #include "s2/s1interval.h"
@@ -54,13 +60,16 @@
 #include "s2/s2point_compression.h"
 #include "s2/s2pointutil.h"
 #include "s2/s2predicates.h"
+#include "s2/s2random.h"
 #include "s2/s2shape.h"
 #include "s2/s2testing.h"
 #include "s2/s2text_format.h"
+#include "s2/util/coding/coder.h"
 #include "s2/util/math/matrix3x3.h"
 
 using absl::flat_hash_map;
 using absl::flat_hash_set;
+using absl::StrCat;
 using absl::string_view;
 using ::s2textformat::MakeLoopOrDie;
 using std::fabs;
@@ -248,7 +257,7 @@ class S2LoopTestBase : public testing::Test {
   }
 
   static const S2Loop* GetLoopIndexPtr(S2Loop& loop) {
-    return down_cast<S2Loop::Shape*>(loop.index_.shape(0))->loop_;
+    return down_cast<const S2Loop::Shape*>(loop.index_.shape(0))->loop_;
   }
 };
 
@@ -309,11 +318,12 @@ TEST_F(S2LoopTestBase, GetAreaConsistentWithSign) {
   // Test that GetArea() returns an area near 0 for degenerate loops that
   // contain almost no points, and an area near 4*Pi for degenerate loops that
   // contain almost all points.
+  absl::BitGen bitgen(S2Testing::MakeTaggedSeedSeq(
+      "GET_AREA_CONSISTENT_WITH_SIGN", absl::LogInfoStreamer(__FILE__, __LINE__).stream()));
 
-  S2Testing::Random* rnd = &S2Testing::rnd;
-  static const int kMaxVertices = 6;
+  static constexpr int kMaxVertices = 6;
   for (int i = 0; i < 50; ++i) {
-    int num_vertices = 3 + rnd->Uniform(kMaxVertices - 3 + 1);
+    int num_vertices = absl::Uniform(bitgen, 3, kMaxVertices + 1);
     // Repeatedly choose N vertices that are exactly on the equator until we
     // find some that form a valid loop.
     S2Loop loop;
@@ -324,7 +334,8 @@ TEST_F(S2LoopTestBase, GetAreaConsistentWithSign) {
         // We limit longitude to the range [0, 90] to ensure that the loop is
         // degenerate (as opposed to following the entire equator).
         vertices.push_back(
-            S2LatLng::FromRadians(0, rnd->RandDouble() * M_PI_2).ToPoint());
+            S2LatLng::FromRadians(0, absl::Uniform(bitgen, 0.0, M_PI_2))
+                .ToPoint());
       }
       loop.Init(vertices);
     } while (!loop.IsValid());
@@ -353,10 +364,12 @@ TEST_F(S2LoopTestBase, GetAreaAndCentroid) {
   // with closely spaces vertices.  Then check that the area and centroid are
   // correct.
 
+  absl::BitGen bitgen(S2Testing::MakeTaggedSeedSeq(
+      "GET_AREA_AND_CENTROID", absl::LogInfoStreamer(__FILE__, __LINE__).stream()));
   for (int i = 0; i < 50; ++i) {
     // Choose a coordinate frame for the spherical cap.
     S2Point x, y, z;
-    S2Testing::GetRandomFrame(&x, &y, &z);
+    s2random::Frame(bitgen, x, y, z);
 
     // Given two points at latitude phi and whose longitudes differ by dtheta,
     // the geodesic between the two points has a maximum latitude of
@@ -367,17 +380,17 @@ TEST_F(S2LoopTestBase, GetAreaAndCentroid) {
     // maximum distance from the boundary of the spherical cap is kMaxDist.
     // Thus we want fabs(atan(tan(phi) / cos(dtheta/2)) - phi) <= kMaxDist.
     static const double kMaxDist = 1e-6;
-    double height = 2 * S2Testing::rnd.RandDouble();
+    double height = absl::Uniform(bitgen, 0.0, 2.0);
     double phi = asin(1 - height);
     double max_dtheta = 2 * acos(tan(fabs(phi)) / tan(fabs(phi) + kMaxDist));
     max_dtheta = min(M_PI, max_dtheta);  // At least 3 vertices.
 
     vector<S2Point> vertices;
     for (double theta = 0; theta < 2 * M_PI;
-         theta += S2Testing::rnd.RandDouble() * max_dtheta) {
-      vertices.push_back(cos(theta) * cos(phi) * x +
-                         sin(theta) * cos(phi) * y +
-                         sin(phi) * z);
+         theta += absl::Uniform(bitgen, 0.0, max_dtheta)) {
+      vertices.push_back(
+          (cos(theta) * cos(phi) * x + sin(theta) * cos(phi) * y + sin(phi) * z)
+              .Normalize());
     }
     S2Loop loop(vertices);
     double area = loop.GetArea();
@@ -427,7 +440,7 @@ TEST_F(S2LoopTestBase, GetCurvature) {
   // to test that the error in GetCurvature is linear in the number of
   // vertices even when the partial sum of the curvatures gets very large.
   // The spiral consists of two "arms" defining opposite sides of the loop.
-  const int kArmPoints = 10000;    // Number of vertices in each "arm"
+  constexpr int kArmPoints = 10000;  // Number of vertices in each "arm"
   const double kArmRadius = 0.01;  // Radius of spiral.
   vector<S2Point> vertices(2 * kArmPoints);
   vertices[kArmPoints] = S2Point(0, 0, 1);
@@ -530,6 +543,11 @@ TEST_F(S2LoopTestBase, Contains) {
   }
 }
 
+TEST(S2Loop, DefaultLoopIsInvalid) {
+  S2Loop loop;
+  EXPECT_FALSE(loop.IsValid());
+}
+
 TEST(S2Loop, ContainsMatchesCrossingSign) {
   // This test demonstrates a former incompatibility between CrossingSign()
   // and Contains(const S2Point&).  It constructs an S2Cell-based loop L and
@@ -872,7 +890,7 @@ static unique_ptr<S2Loop> MakeCellLoop(S2CellId begin, S2CellId end) {
   vector<S2Point> vertices;
   S2Point p = edges.begin()->first;
   while (!edges.empty()) {
-    S2_DCHECK_EQ(1, edges[p].size());
+    ABSL_DCHECK_EQ(1, edges[p].size());
     S2Point next = *edges[p].begin();
     vertices.push_back(p);
     edges.erase(p);
@@ -886,15 +904,16 @@ TEST(S2Loop, LoopRelations2) {
   // Construct polygons consisting of a sequence of adjacent cell ids
   // at some fixed level.  Comparing two polygons at the same level
   // ensures that there are no T-vertices.
+  absl::BitGen bitgen(S2Testing::MakeTaggedSeedSeq(
+      "LOOP_RELATIONS2", absl::LogInfoStreamer(__FILE__, __LINE__).stream()));
   for (int iter = 0; iter < 1000; ++iter) {
-    S2Testing::Random& rnd = S2Testing::rnd;
-    S2CellId begin = S2CellId(rnd.Rand64() | 1);
+    S2CellId begin = S2CellId(absl::Uniform<uint64_t>(bitgen) | 1);
     if (!begin.is_valid()) continue;
-    begin = begin.parent(rnd.Uniform(S2CellId::kMaxLevel));
-    S2CellId a_begin = begin.advance(rnd.Skewed(6));
-    S2CellId a_end = a_begin.advance(rnd.Skewed(6) + 1);
-    S2CellId b_begin = begin.advance(rnd.Skewed(6));
-    S2CellId b_end = b_begin.advance(rnd.Skewed(6) + 1);
+    begin = begin.parent(absl::Uniform(bitgen, 0, S2CellId::kMaxLevel));
+    S2CellId a_begin = begin.advance(s2random::SkewedInt(bitgen, 6));
+    S2CellId a_end = a_begin.advance(s2random::SkewedInt(bitgen, 6) + 1);
+    S2CellId b_begin = begin.advance(s2random::SkewedInt(bitgen, 6));
+    S2CellId b_end = b_begin.advance(s2random::SkewedInt(bitgen, 6) + 1);
     if (!a_end.is_valid() || !b_end.is_valid()) continue;
 
     unique_ptr<S2Loop> a(MakeCellLoop(a_begin, a_end));
@@ -902,13 +921,13 @@ TEST(S2Loop, LoopRelations2) {
     if (a.get() && b.get()) {
       bool contained = (a_begin <= b_begin && b_end <= a_end);
       bool intersects = (a_begin < b_end && b_begin < a_end);
-      S2_VLOG(1) << "Checking " << a->num_vertices() << " vs. "
-              << b->num_vertices() << ", contained = " << contained
-              << ", intersects = " << intersects;
+      ABSL_VLOG(1) << "Checking " << a->num_vertices() << " vs. "
+                   << b->num_vertices() << ", contained = " << contained
+                   << ", intersects = " << intersects;
       EXPECT_EQ(a->Contains(*b.get()), contained);
       EXPECT_EQ(a->Intersects(*b.get()), intersects);
     } else {
-      S2_VLOG(1) << "MakeCellLoop failed to create a loop.";
+      ABSL_VLOG(1) << "MakeCellLoop failed to create a loop.";
     }
   }
 }
@@ -917,16 +936,17 @@ TEST(S2Loop, BoundsForLoopContainment) {
   // To reliably test whether one loop contains another, the bounds of the
   // outer loop are expanded slightly.  This test constructs examples where
   // this expansion is necessary and verifies that it is sufficient.
+  absl::BitGen bitgen(S2Testing::MakeTaggedSeedSeq(
+      "BOUNDS_FOR_LOOP_CONTAINMENT", absl::LogInfoStreamer(__FILE__, __LINE__).stream()));
 
-  S2Testing::Random* rnd = &S2Testing::rnd;
   for (int iter = 0; iter < 1000; ++iter) {
     // We construct a triangle ABC such that A,B,C are nearly colinear, B is
     // the point of maximum latitude, and the edge AC passes very slightly
     // below B (i.e., ABC is CCW).
-    S2Point b = (S2Testing::RandomPoint() + S2Point(0, 0, 1)).Normalize();
+    S2Point b = (s2random::Point(bitgen) + S2Point(0, 0, 1)).Normalize();
     S2Point v = b.CrossProd(S2Point(0, 0, 1)).Normalize();
-    S2Point a = S2::Interpolate(-v, b, rnd->RandDouble());
-    S2Point c = S2::Interpolate(b, v, rnd->RandDouble());
+    S2Point a = S2::Interpolate(-v, b, absl::Uniform(bitgen, 0.0, 1.0));
+    S2Point c = S2::Interpolate(b, v, absl::Uniform(bitgen, 0.0, 1.0));
     if (s2pred::Sign(a, b, c) < 0) {
       --iter; continue;
     }
@@ -1045,7 +1065,7 @@ TEST(S2Loop, Moveable) {
 }
 
 static void TestEmptyFullSnapped(const S2Loop& loop, int level) {
-  S2_CHECK(loop.is_empty_or_full());
+  ABSL_CHECK(loop.is_empty_or_full());
   S2CellId cellid = S2CellId(loop.vertex(0)).parent(level);
   vector<S2Point> vertices = {cellid.ToPoint()};
   S2Loop loop2(vertices);
@@ -1057,7 +1077,7 @@ static void TestEmptyFullSnapped(const S2Loop& loop, int level) {
 // Test converting the empty/full loops to S2LatLng representations.  (We
 // don't bother testing E5/E6/E7 because that test is less demanding.)
 static void TestEmptyFullLatLng(const S2Loop& loop) {
-  S2_CHECK(loop.is_empty_or_full());
+  ABSL_CHECK(loop.is_empty_or_full());
   vector<S2Point> vertices = {S2LatLng(loop.vertex(0)).ToPoint()};
   S2Loop loop2(vertices);
   EXPECT_TRUE(loop.BoundaryEquals(loop2));
@@ -1135,15 +1155,15 @@ static void CheckLoopIsInvalid(string_view str, string_view snippet) {
   unique_ptr<S2Loop> loop(MakeLoopOrDie(str, S2Debug::DISABLE));
   S2Error error;
   EXPECT_TRUE(loop->FindValidationError(&error));
-  EXPECT_THAT(error.text(), testing::HasSubstr(snippet));
+  EXPECT_THAT(error.message(), testing::HasSubstr(snippet));
 }
 
-static void CheckLoopIsInvalid(const vector<S2Point>& points,
+static void CheckLoopIsInvalid(absl::Span<const S2Point> points,
                                string_view snippet) {
   S2Loop l(points, S2Debug::DISABLE);
   S2Error error;
   EXPECT_TRUE(l.FindValidationError(&error));
-  EXPECT_THAT(error.text(), testing::HasSubstr(snippet));
+  EXPECT_THAT(error.message(), testing::HasSubstr(snippet));
 }
 
 TEST(S2Loop, IsValidDetectsInvalidLoops) {
@@ -1179,7 +1199,7 @@ TEST_F(S2LoopTestBase, PointersCorrectAfterMove) {
 
 #if GTEST_HAS_DEATH_TEST
 TEST(S2LoopDeathTest, IsValidDetectsInvalidLoops) {
-  // Points with length > sqrt(2) (triggers S2_DCHECK failure in debug)
+  // Points with length > sqrt(2) (triggers ABSL_DCHECK failure in debug)
   EXPECT_DEBUG_DEATH(
       CheckLoopIsInvalid({S2Point(2, 0, 0), S2Point(0, 1, 0), S2Point(0, 0, 1)},
                          "unit length"),
@@ -1330,3 +1350,8 @@ TEST(S2LoopOwningShape, Ownership) {
   S2Loop::OwningShape shape(std::move(loop));
 }
 
+TEST(S2LoopOwningShape, Init) {
+  S2Loop::OwningShape shape;
+  shape.Init(make_unique<S2Loop>(S2Loop::kEmpty()));
+}
+
