@@ -29,14 +29,12 @@
 // Avoid adding expensive includes here.
 #include "absl/base/attributes.h"
 #include "absl/base/macros.h"
+#include "absl/base/optimization.h"
+#include "absl/log/absl_check.h"
 #include "absl/meta/type_traits.h"
 #include "absl/numeric/int128.h"
-#include "absl/utility/utility.h"
+#include "absl/strings/string_view.h"
 
-#include "s2/base/casts.h"
-#include "s2/base/integral_types.h"
-#include "s2/base/logging.h"
-#include "s2/base/port.h"
 #include "s2/util/coding/varint.h"
 #include "s2/util/endian/endian.h"
 
@@ -60,34 +58,90 @@ class Encoder {
   void reset(void* buf, size_t maxn);
   void clear();
 
-  // Encoding routines.  Note that these do not check bounds
-  void put8(unsigned char v) { writer().put8(v); }
-  void put16(uint16 v) { writer().put16(v); }
-  void put32(uint32 v) { writer().put32(v); }
-  void put64(uint64 v) { writer().put64(v); }
-  void put128(absl::uint128 v) { writer().put128(v); }
-  void putn(const void* mem, size_t n) { writer().putn(mem, n); }
+  // Encoding routines. Callers must know that the remaining buffer has
+  // sufficient size.
+  void put8(unsigned char v) {
+    ABSL_HARDENING_ASSERT(avail() >= sizeof(v));
+    writer().put8(v);
+  }
+  void put16(uint16_t v) {
+    ABSL_HARDENING_ASSERT(avail() >= sizeof(v));
+    writer().put16(v);
+  }
+  void put32(uint32_t v) {
+    ABSL_HARDENING_ASSERT(avail() >= sizeof(v));
+    writer().put32(v);
+  }
+  void put64(uint64_t v) {
+    ABSL_HARDENING_ASSERT(avail() >= sizeof(v));
+    writer().put64(v);
+  }
+  void put128(absl::uint128 v) {
+    ABSL_HARDENING_ASSERT(avail() >= sizeof(v));
+    writer().put128(v);
+  }
+  void putn(const void* mem, size_t n) {
+    ABSL_HARDENING_ASSERT(avail() >= n);
+    writer().putn(mem, n);
+  }
 
   // Put no more than n bytes, stopping when c is put.
-  void putcn(const void* mem, int c, size_t n) { writer().putcn(mem, c, n); }
+  void putcn(const void* mem, int c, size_t n) {
+    ABSL_HARDENING_ASSERT(avail() >= n);
+    writer().putcn(mem, c, n);
+  }
 
-  // put a c-string including \0.
-  void puts(const void* mem) { writer().puts(mem); }
-  // put a c-string without \0.
-  void puts_without_null(const char* mem) { writer().puts_no_null(mem); }
-  void putfloat(float f) { writer().put32(absl::bit_cast<uint32>(f)); }
-  void putdouble(double d) { writer().put64(absl::bit_cast<uint64>(d)); }
+  // Put a c-string including \0.
+  void puts(const void* mem) {
+    // No ABSL_HARDENING_ASSERT. This eventually calls memccpy which has size
+    // parameter.
+    writer().puts(mem);
+  }
+
+  // Put the contents of c-string up to but not including the terminating NUL
+  // character.
+  // `mem` must not be nullptr.
+  // `mem` must be a NUL-terminated string.
+  void puts_without_null(const char* mem) {
+    // No ABSL_HARDENING_ASSERT. The implementation stops writing at the end of
+    // the buffer.
+    writer().puts_no_null(mem);
+  }
+
+  void putfloat(float f) {
+    ABSL_HARDENING_ASSERT(avail() >= sizeof(uint32_t));
+    writer().put32(absl::bit_cast<uint32_t>(f));
+  }
+  void putdouble(double d) {
+    ABSL_HARDENING_ASSERT(avail() >= sizeof(uint64_t));
+    writer().put64(absl::bit_cast<uint64_t>(d));
+  }
 
   // Support for variable length encoding with 7 bits per byte
   // (these are just simple wrappers around the Varint module)
   static constexpr int kVarintMax32 = Varint::kMax32;
   static constexpr int kVarintMax64 = Varint::kMax64;
 
-  void put_varint32(uint32 v) { writer().put_varint32(v); }
-  void put_varint32_inline(uint32 v) { writer().put_varint32_inline(v); }
-  void put_varint64(uint64 v) { writer().put_varint64(v); }
-  static int varint32_length(uint32 v);  // Length of var encoding of "v"
-  static int varint64_length(uint64 v);  // Length of var encoding of "v"
+  void put_varint32(uint32_t v) {
+    // The majority of the time we can avoid computing Length32().
+    ABSL_HARDENING_ASSERT(avail() >= static_cast<size_t>(Varint::kMax32) ||
+                          avail() >= static_cast<size_t>(Varint::Length32(v)));
+    writer().put_varint32(v);
+  }
+  void put_varint32_inline(uint32_t v) {
+    // The majority of the time we can avoid computing Length32().
+    ABSL_HARDENING_ASSERT(avail() >= static_cast<size_t>(Varint::kMax32) ||
+                          avail() >= static_cast<size_t>(Varint::Length32(v)));
+    writer().put_varint32_inline(v);
+  }
+  void put_varint64(uint64_t v) {
+    // The majority of the time we can avoid computing Length64().
+    ABSL_HARDENING_ASSERT(avail() >= static_cast<size_t>(Varint::kMax64) ||
+                          avail() >= static_cast<size_t>(Varint::Length64(v)));
+    writer().put_varint64(v);
+  }
+  static int varint32_length(uint32_t v);  // Length of var encoding of "v"
+  static int varint64_length(uint64_t v);  // Length of var encoding of "v"
 
   // The fast implementation of the code below with boundary checks.
   //   uint64 val;
@@ -131,11 +185,19 @@ class Encoder {
 
   // Advances the write pointer by "N" bytes. It returns the position of the
   // pointer before the skip (in other words start of the skipped bytes).
-  char* skip(ptrdiff_t N) { return writer().skip(N); }
+  char* skip(ptrdiff_t N) {
+    // Negative values of N are allowed. Ensure the pointer lands within the
+    // original bounds.
+    ABSL_HARDENING_ASSERT(static_cast<ptrdiff_t>(length()) + N >= 0 &&
+                          static_cast<ptrdiff_t>(avail()) >= N);
+    return writer().skip(N); }
 
   // REQUIRES: length() >= N
   // Removes the last N bytes out of the encoded buffer
-  void RemoveLast(size_t N) { writer().skip(-static_cast<ptrdiff_t>(N)); }
+  void RemoveLast(size_t N) {
+    ABSL_HARDENING_ASSERT(length() >= N);
+    writer().skip(-static_cast<ptrdiff_t>(N));
+  }
 
   // REQUIRES: length() >= N
   // Removes the last length()-N bytes to make the encoded buffer have length N
@@ -144,26 +206,23 @@ class Encoder {
  private:
   // All encoding operations are done through the Writer. This avoids aliasing
   // between `buf_` and `this` which allows the compiler to avoid reloading
-  // `buf_` repeatedly. See https://godbolt.org/z/zM36s3ded.
+  // `buf_` repeatedly. See https://godbolt.org/z/sfEnePYWP.
   struct Writer {
     Encoder* enc;
-    char* p;
-
-    explicit Writer(Encoder* e)
-        : enc(e), p(reinterpret_cast<char*>(enc->buf_)) {}
+    char* p{reinterpret_cast<char*>(enc->buf_)};
 
     ~Writer() {
       enc->buf_ = reinterpret_cast<unsigned char*>(p);
-      S2_DCHECK_GE(enc->buf_, enc->orig_);
-      S2_DCHECK_LE(enc->buf_, enc->limit_);
+      ABSL_DCHECK_GE(enc->buf_, enc->orig_);
+      ABSL_DCHECK_LE(enc->buf_, enc->limit_);
     }
 
-    char* skip(ptrdiff_t N) { return absl::exchange(p, p + N); }
+    char* skip(ptrdiff_t N) { return std::exchange(p, p + N); }
 
     void put8(unsigned char v) { *p++ = v; }
-    void put16(uint16 v) { LittleEndian::Store16(skip(2), v); }
-    void put32(uint32 v) { LittleEndian::Store32(skip(4), v); }
-    void put64(uint64 v) { LittleEndian::Store64(skip(8), v); }
+    void put16(uint16_t v) { LittleEndian::Store16(skip(2), v); }
+    void put32(uint32_t v) { LittleEndian::Store32(skip(4), v); }
+    void put64(uint64_t v) { LittleEndian::Store64(skip(8), v); }
     void put128(absl::uint128 v) { LittleEndian::Store128(skip(16), v); }
     void putn(const void* src, size_t n) { memcpy(skip(n), src, n); }
     void putcn(const void* src, int c, size_t n) {
@@ -176,18 +235,18 @@ class Encoder {
       auto* l = reinterpret_cast<char*>(enc->limit_);
       while (*mem != '\0' && p < l) *p++ = *mem++;
     }
-    ABSL_ATTRIBUTE_ALWAYS_INLINE void put_varint32(uint32 v) {
+    ABSL_ATTRIBUTE_ALWAYS_INLINE void put_varint32(uint32_t v) {
       p = Varint::Encode32(p, v);
     }
-    ABSL_ATTRIBUTE_ALWAYS_INLINE void put_varint32_inline(uint32 v) {
+    ABSL_ATTRIBUTE_ALWAYS_INLINE void put_varint32_inline(uint32_t v) {
       p = Varint::Encode32Inline(p, v);
     }
-    ABSL_ATTRIBUTE_ALWAYS_INLINE void put_varint64(uint64 v) {
+    ABSL_ATTRIBUTE_ALWAYS_INLINE void put_varint64(uint64_t v) {
       p = Varint::Encode64(p, v);
     }
   };
 
-  Writer writer() { return Writer(this); }
+  Writer writer() { return Writer{this}; }
 
   static std::pair<unsigned char*, size_t> NewBuffer(size_t size);
   static void DeleteBuffer(unsigned char* buf, size_t size);
@@ -220,7 +279,7 @@ class Encoder {
 class Decoder {
  public:
   // Empty constructor to create uninitialized decoder
-  inline Decoder() { }
+  Decoder() = default;
 
   // NOTE: for efficiency reasons, this is not virtual.  so don't add
   // any members that really need to be destructed, and be careful about
@@ -229,14 +288,16 @@ class Decoder {
   // ~Decoder() = default;
 
   // Initialize decoder to decode from "buf"
+  explicit Decoder(absl::string_view buf) : Decoder(buf.data(), buf.size()) {}
   Decoder(const void* buf, size_t maxn);
   void reset(const void* buf, size_t maxn);
 
-  // Decoding routines.  Note that these do not check bounds
+  // Decoding routines.  Callers must know that the remaining buffer has
+  // sufficient size.
   unsigned char get8();
-  uint16 get16();
-  uint32 get32();
-  uint64 get64();
+  uint16_t get16();
+  uint32_t get32();
+  uint64_t get64();
   absl::uint128 get128();
   float  getfloat();
   double getdouble();
@@ -250,8 +311,8 @@ class Decoder {
   unsigned char const* ptr() const;  // Return ptr to current position in buffer
 
   // "get_varint" actually checks bounds
-  bool get_varint32(uint32* v);
-  bool get_varint64(uint64* v);
+  bool get_varint32(uint32_t* v);
+  bool get_varint64(uint64_t* v);
 
   size_t pos() const;
   // Return number of bytes decoded so far
@@ -302,20 +363,20 @@ inline void Encoder::clear() {
 }
 
 inline void Encoder::Ensure(size_t N) {
-  S2_DCHECK(ensure_allowed());
+  ABSL_DCHECK(ensure_allowed());
   if (avail() < N) {
     EnsureSlowPath(N);
   }
 }
 
 inline size_t Encoder::length() const {
-  S2_DCHECK_GE(buf_, orig_);
-  S2_DCHECK_LE(buf_, limit_);
+  ABSL_DCHECK_GE(buf_, orig_);
+  ABSL_DCHECK_LE(buf_, limit_);
   return buf_ - orig_;
 }
 
 inline size_t Encoder::avail() const {
-  S2_DCHECK_GE(limit_, buf_);
+  ABSL_DCHECK_GE(limit_, buf_);
   return limit_ - buf_;
 }
 
@@ -323,7 +384,7 @@ inline size_t Encoder::avail() const {
 // Template parameter N specifies the number of bytes to copy. Passing
 // constant size results in optimized code from memcpy for the size.
 template <size_t N>
-void CopyAndAdvance(const uint8** src, uint8** dst) {
+void CopyAndAdvance(const uint8_t** src, uint8_t** dst) {
   memcpy(*dst, *src, N);
   *dst += N;
   *src += N;
@@ -422,7 +483,7 @@ ABSL_ATTRIBUTE_ALWAYS_INLINE inline bool Encoder::put_varint64_from_decoder(
     if (ABSL_PREDICT_FALSE(enc_ptr >= limit_ - 1)) {
       return false;
     }
-    UNALIGNED_STORE16(enc_ptr, UNALIGNED_LOAD16(dec_ptr));
+    memcpy(enc_ptr, dec_ptr, 2);
     dec->buf_ += 2;
     buf_ += 2;
     return true;
@@ -442,12 +503,12 @@ inline void Decoder::reset(const void* b, size_t maxn) {
 }
 
 inline size_t Decoder::pos() const {
-  S2_DCHECK_GE(buf_, orig_);
+  ABSL_DCHECK_GE(buf_, orig_);
   return buf_ - orig_;
 }
 
 inline size_t Decoder::avail() const {
-  S2_DCHECK_GE(limit_, buf_);
+  ABSL_DCHECK_GE(limit_, buf_);
   return limit_ - buf_;
 }
 
@@ -468,7 +529,7 @@ inline void Decoder::getcn(void* dst, int c, size_t n) {
 
 inline void Decoder::gets(void* dst, size_t n) {
   size_t len = n - 1;
-  S2_DCHECK_GE(limit_, buf_);
+  ABSL_DCHECK_GE(limit_, buf_);
   if (n > static_cast<size_t>(1 + limit_ - buf_)) {
     len = limit_ - buf_;
   }
@@ -477,6 +538,10 @@ inline void Decoder::gets(void* dst, size_t n) {
 }
 
 inline const char* Decoder::skip(ptrdiff_t n) {
+  // Negative values of n are allowed. Ensure the pointer lands within the
+  // original bounds.
+  ABSL_HARDENING_ASSERT(static_cast<ptrdiff_t>(avail()) >= n &&
+                        static_cast<ptrdiff_t>(pos()) + n >= 0);
   auto* start = reinterpret_cast<const char*>(buf_);
   buf_ += n;
   return start;
@@ -500,30 +565,35 @@ inline void DecoderExtensions::FillArray(Decoder* array, int num_decoders) {
 }
 
 inline unsigned char Decoder::get8() {
+  ABSL_HARDENING_ASSERT(avail() >= sizeof(unsigned char));
   const unsigned char v = *buf_;
   buf_ += sizeof(v);
   return v;
 }
 
-inline uint16 Decoder::get16() {
-  const uint16 v = LittleEndian::Load16(buf_);
+inline uint16_t Decoder::get16() {
+  ABSL_HARDENING_ASSERT(avail() >= sizeof(uint16_t));
+  const uint16_t v = LittleEndian::Load16(buf_);
   buf_ += sizeof(v);
   return v;
 }
 
-inline uint32 Decoder::get32() {
-  const uint32 v = LittleEndian::Load32(buf_);
+inline uint32_t Decoder::get32() {
+  ABSL_HARDENING_ASSERT(avail() >= sizeof(uint32_t));
+  const uint32_t v = LittleEndian::Load32(buf_);
   buf_ += sizeof(v);
   return v;
 }
 
-inline uint64 Decoder::get64() {
-  const uint64 v = LittleEndian::Load64(buf_);
+inline uint64_t Decoder::get64() {
+  ABSL_HARDENING_ASSERT(avail() >= sizeof(uint64_t));
+  const uint64_t v = LittleEndian::Load64(buf_);
   buf_ += sizeof(v);
   return v;
 }
 
 inline absl::uint128 Decoder::get128() {
+  ABSL_HARDENING_ASSERT(avail() >= sizeof(absl::uint128));
   const absl::uint128 v = LittleEndian::Load128(buf_);
   buf_ += sizeof(v);
   return v;
@@ -537,7 +607,7 @@ inline double Decoder::getdouble() {
   return absl::bit_cast<double>(get64());
 }
 
-inline bool Decoder::get_varint32(uint32* v) {
+inline bool Decoder::get_varint32(uint32_t* v) {
   const char* const r =
       Varint::Parse32WithLimit(reinterpret_cast<const char*>(buf_),
                                reinterpret_cast<const char*>(limit_), v);
@@ -548,7 +618,7 @@ inline bool Decoder::get_varint32(uint32* v) {
   return true;
 }
 
-inline bool Decoder::get_varint64(uint64* v) {
+inline bool Decoder::get_varint64(uint64_t* v) {
   const char* const r =
       Varint::Parse64WithLimit(reinterpret_cast<const char*>(buf_),
                                reinterpret_cast<const char*>(limit_), v);
